#' Density of a dimension-wise shifted exponential normal.
#'
#' @param x Data Matrix.
#' @param mu Mean vector.
#' @param Sigma Scale matrix.
#' @param theta Vector of tailedness parameters for each dimension.
#'
#' @return The value(s) of the density in \code{x}.
#' @examples
#' d <- 3
#' x <- matrix(rnorm(d*2), 2, d)
#' dDSSEN(x, mu = rep(0,d), Sigma = diag(d), theta = rep(1,d))
#'
#' @export
dDSSEN <- function(x, mu = rep(0,d), Sigma, theta = Inf){

  if(missing(Sigma))
    stop("Sigma is missing")
  if(sum(theta < 0)>0)
    stop("theta must be greater than, or equal to, 0")

  if(is.matrix(Sigma)){
    d <- ncol(Sigma)
    q <- nrow(x)
  }

  if(!is.matrix(Sigma)){
    d <- 1
    q <- length(x)
  }

  if(is.vector(x)){
    x <- matrix(x,length(x),1)
    Sigma <- matrix(Sigma,nrow=d,ncol=d)
  }

  PDF <- NULL

  Sigmainv <- solve(Sigma)
  const1   <- 2^d/(sqrt(det(Sigma))) * prod(theta) * prod(exp(theta))

  if(d==1){

    Dt       <- theta

    fpass <- function(z){

      Dy     <- z-mu
      P      <- Dy %*% Sigmainv %*% Dy
      Pt     <- P + 2*Dt
      Ptinv  <- solve(Pt)
      const2 <- sqrt(det(Ptinv))

      M <- as.vector(mnormt::recintab(rep(2,d), a=rep(1,d), b=rep(Inf,d), mu=rep(0,d), S=Ptinv))

      PDFpass <- const1 * const2 * M[length(M)]

      return(PDFpass)

    }
  }

  if(d>1){

    Dt       <- diag(theta)

    fpass <- function(z){

      Dy     <- diag(z-mu)
      P      <- Dy %*% Sigmainv %*% Dy
      Pt     <- P + 2*Dt
      Ptinv  <- solve(Pt)
      const2 <- sqrt(det(Ptinv))

      M <- as.vector(mnormt::recintab(rep(2,d), a=rep(1,d), b=rep(Inf,d), mu=rep(0,d), S=Ptinv))

      PDFpass <- const1 * const2 * M[length(M)]

      return(PDFpass)

    }
  }

  PDF <- sapply(1:q, function(i) fpass(z=x[i,]) )

  return(PDF)

}
