#' Density of a dimension-wise tail-inflated normal.
#'
#' @param x Data Matrix.
#' @param mu Mean vector.
#' @param Sigma Scale matrix.
#' @param theta Vector of tailedness parameters for each dimension.
#'
#' @return The value(s) of the density in \code{x}.
#' @examples
#' d <- 3
#' x <- matrix(rnorm(d*2), 2, d)
#' dDSTIN(x, mu = rep(0,d), Sigma = diag(d), theta = rep(0.4,d))
#'
#' @export
dDSTIN <- function(x, mu = rep(0,d), Sigma, theta){

  if(missing(Sigma))
    stop("Sigma is missing")

  if(min(theta) <= 0 || max(theta) >= 1)
    stop("each element in theta must be in (0,1)")

  if(is.matrix(Sigma)){
    d <- ncol(Sigma)
    q <- nrow(x)
  }

  if(!is.matrix(Sigma)){
    d <- 1
    q <- length(x)
  }

  if(is.vector(x)){
    x <- matrix(x,length(x),1)
    Sigma <- matrix(Sigma,nrow=d,ncol=d)
  }

  PDF <- NULL

  Sigmainv <- solve(Sigma)
  const1   <- 2^d/prod(theta)

  if(d==1){

    fpass <- function(z){

      Dy     <- (z-mu)
      P      <- Dy %*% Sigmainv %*% Dy
      Pinv   <- solve(P)
      const2 <- 1/abs(Dy)

      M   <- as.vector(mnormt::recintab(rep(2,d), a=sqrt(1-theta), b=rep(1,d), mu=rep(0,d), S=Pinv))

      pez <- M[length(M)]

      if(pez < 0 || abs(z-mu)<0.0001){
        const2 <- (2*pi)^(-1/2)/sqrt(Sigma)
        pez    <- 1/3*(1-(1-theta)^(3/2))
      }

      PDFpass <- const1 * const2 * pez

      return(PDFpass)

    }

  }

  if(d>1){

    fpass <- function(z){

      Dy     <- diag(z-mu)
      P      <- Dy %*% Sigmainv %*% Dy
      Pinv   <- MASS::ginv(P)
      const2 <- 1/(det(Sigma)^(1/2)*det(P)^(1/2))

      #print(Pinv)

      M      <- as.vector(mnormt::recintab(rep(2,d), a=sqrt(1-theta), b=rep(1,d), mu=rep(0,d), S=Pinv))

      pez    <- M[length(M)]
      #print(pez)

      if(is.nan(M[length(M)]) | M[length(M)]<0 | sum(abs(z-mu)<0.0003)>0 ){

        ind <- which(abs(z-mu)==min(abs(z-mu)))

        #print(ind)

        Dy     <- diag(z-mu)
        Ps     <- Dy[-ind,-ind] %*% Sigmainv[-ind,-ind] %*% Dy[-ind,-ind]
        Psinv  <- MASS::ginv(Ps)
        thetas <- theta[-ind]

        const2 <- 1/(det(Sigma)^(1/2)*det(Ps)^(1/2))

        M       <- as.vector(mnormt::recintab(rep(2,(d - length(ind))), a=sqrt(1-thetas), b=rep(1,(d - length(ind))), mu=rep(0,(d - length(ind))), S=Psinv))
        pez1    <- M[length(M)]

        pez2    <- (2*pi)^(-length(ind)/2)*prod(1/3*(1-(1-theta[ind])^(3/2)))

        pez     <- pez1*pez2

        #print("Attention M")

        #pez <- 1e-310


      }

      PDFpass <- const1 * const2 * pez

      PDFmax  <- const1 * 1/det(Sigma)^(1/2) * (2*pi)^(-d/2)*prod(1/3*(1-(1-theta)^(3/2)))

      if(PDFpass > PDFmax)   PDFpass <- PDFmax

      if(PDFpass < 1.0e-323) PDFpass <- 1.0e-323

      return(PDFpass)

    }

  }

  PDF <- sapply(1:q, function(i) fpass(z=x[i,]) )

  return(PDF)

}
