#' ML fitting for dimension wise distributions: EM-based algorithms.
#'
#' @param X Data matrix.
#' @param model marginal distribution considered. Possible values are: \code{"sen"} or \code{"tin"}.
#' @param common.theta boolean indicating if the taildeness parameters are tied (\code{TRUE}) or no (\code{FALSE}) across dimensions.
#' @param method.theta numerical methods used by \code{\link[stats]{optim}} to update \code{theta} in the second CM-step of the ECME algorithm.
#' @param tol Convergence tolerance.
#' @param maxit Maximum number of iterations for the EM-based algorithm.
#' @param trace.theta Non-negative integer. If positive, tracing information on the progress of the optimization is produced. Higher values may produce more tracing information (see \code{\link[stats]{optim}} for further details).
#' @param package The R package used to compute the incomplete gamma function.
#' @param mu.TRUE Mean vector specified by the user.
#' @param Sigma.TRUE Scale matrix specified by the user.
#' @param theta.TRUE Tailedness parameter vector specified by the user.
#'
#' @return A list with the following elements:
#' \item{X}{Data matrix}
#' \item{n}{Sample size.}
#' \item{d}{Dimension.}
#' \item{npar}{Number of (free) parameters.}
#' \item{mu.init}{Mean vector estimated by the method of weighted moments.}
#' \item{Sigma.init}{Scale matrix estimated by the method of weighted moments.}
#' \item{theta.init}{Taildeness parameter vector estimated by the method of weighted moments.}
#' \item{mu}{Mean vector.}
#' \item{Sigma}{Scalar matrix.}
#' \item{theta}{Taildeness parameter vector.}
#' \item{w}{A posteriori mixing weights.}
#' \item{density}{Densities values in correspondence of \code{X}.}
#' \item{iter.stop}{Number of iterations  at convergence.}
#' \item{loglik}{Log-likelihood at convergence.}
#' \item{AIC}{AIC.}
#' \item{BIC}{BIC.}
#' @examples
#' d <- 2
#' set.seed(1234)
#' X <- rDSSEN(100,mu=rep(0,d),Sigma=diag(d),theta=rep(0.01,d))$X
#' dsdist.ML.EM(X = X)
#'
#' @export
dsdist.ML.EM <- function(X,
                         model = "sen",
                         common.theta = FALSE,
                         algorithm = "EM",
                         initialization = "WMM", #  "MN"
                         method.theta = "Nelder-Mead",
                         tol = 10^-5,
                         maxit = 1000,
                         trace.theta = 0,
                         package = "zipfR",
                         mu.TRUE = NULL,
                         Sigma.TRUE = NULL,
                         theta.TRUE = NULL){

  if(is.vector(X))
    X <- matrix(X, ncol = 1)
  if(is.data.frame(X))
    X <- as.matrix(X)
  if(!is.matrix(X))
    X <- matrix(X, nrow = length(X), ncol = 1, dimnames = list(names(X), deparse(substitute(X))))
  if(!is.numeric(X))
    stop("numeric matrix/vector expected for X")
  if(any(is.na(X)))
    stop("No NAs allowed")
  if(algorithm == "EM" & model == "tin"){
    warning("EM has been replaced by the ECME because for the DSTIN distribution the EM algorithm is not well-defined")
    algorithm <- "ECME"
  }

  n <- nrow(X) # sample size
  d <- ncol(X) # number of variables

  # Fixing some of the parameters

  if(!is.null(mu.TRUE)){
    mu <- mu.TRUE
  }
  if(!is.null(Sigma.TRUE)){
    Sigma <- Sigma.TRUE
  }
  if(!is.null(theta.TRUE)){
    theta <- theta.TRUE
  }

  # ------------------------------------------- #
  # Initialization of the M-step (or CM-step 1) #
  # ------------------------------------------- #

  # Method of moments

  if(initialization == "WMM"){

    temp       <- dsdist_WMM(X = X, model = model, weights = NULL)
    mu.init    <- temp$mu
    Sigma.init <- temp$Sigma
    theta.init <- temp$theta

  }

  # Multivariate normal

  if(initialization == "MN"){

    temp       <- Rfast::mvnorm.mle(X)
    mu.init    <- temp$mu
    Sigma.init <- temp$sigma
    if(model=="sen")
      theta.init <- rep(345,d)
    if(model=="tin")
      theta.init <- rep(0.01,d)

  }

  if(is.null(mu.TRUE)){
    mu <- mu.init
  }
  if(is.null(Sigma.TRUE)){
    Sigma <- Sigma.init
  }
  if(is.null(theta.TRUE)){
    if(!common.theta)
      theta <- theta.init
    if(common.theta)
      theta <- rep(mean(theta.init),d)
  }

  mu.rep    <- matrix(rep(mu,n),n,d,byrow=TRUE)
  Sigma.rep <- matrix(rep(ifelse(d>1,diag(Sigma),Sigma),n),n,d,byrow=TRUE)

  # Definitions of objects for the E-step

  w     <- array(NA,c(n,d),dimnames=list(1:n,paste("dim.",1:d,sep="")))
  delta <- array(NA,c(n,d),dimnames=list(1:n,paste("dim.",1:d,sep="")))

  # Preliminary definition of convergence criterions

  check     <- 0
  iteration <- 1
  loglik    <- -Inf

  while(check<1){

    # ------ #
    # E-step #
    # ------ #

    w.old <- w
    delta <- (X-mu.rep)^2/Sigma.rep
    for(i in 1:n){

      if(model == "sen"){
        for(h in 1:d){

          num    <- expint::gammainc(a = (1/2 + 2), x = (delta[i,h]/2 + theta[h]))
          den    <- (delta[i,h]/2 + theta[h])*expint::gammainc(a = (1/2 + 1), x = (delta[i,h]/2 + theta[h]))
          w[i,h] <- num/den

        }
      }
      if(model == "tin"){
        for(h in 1:d){

          if(package == "zipfR"){
            #num  <-          max(2*(zipfR::Igamma(a=(1/2+2), x=(1-theta[h])*delta[i,h]/2, lower=FALSE) - zipfR::Igamma(a=(1/2+2), x=delta[i,h]/2, lower=FALSE)),10^(-322))
            #den  <- max(delta[i,h]*(zipfR::Igamma(a=(1/2+1), x=(1-theta[h])*delta[i,h]/2, lower=FALSE) - zipfR::Igamma(a=(1/2+1), x=delta[i,h]/2, lower=FALSE)),10^(-322))
            num  <-          2*(zipfR::Igamma(a=(1/2+2), x=(1-theta[h])*delta[i,h]/2, lower=FALSE) - zipfR::Igamma(a=(1/2+2), x=delta[i,h]/2, lower=FALSE))
            den  <- delta[i,h]*(zipfR::Igamma(a=(1/2+1), x=(1-theta[h])*delta[i,h]/2, lower=FALSE) - zipfR::Igamma(a=(1/2+1), x=delta[i,h]/2, lower=FALSE))
          }
          if(package == "expint"){
            num    <-          2*abs(expint::gammainc(a=(1/2+2), x=(1-theta[h])*delta[i,h]/2) - expint::gammainc(a=(1/2+2), x=delta[i,h]/2))
            den    <- delta[i,h]*abs(expint::gammainc(a=(1/2+1), x=(1-theta[h])*delta[i,h]/2) - expint::gammainc(a=(1/2+1), x=delta[i,h]/2))
          }
          w[i,h] <- num/den

          if(is.nan(w[i,h]) | w[i,h] >= 1) # delta -> 0 menas x_ih -> mu_h
            w[i,h] <- 0.999
          if(w[i,h] <= (1-theta[h]))
            w[i,h] <- (1-theta[h]) + 0.001

        }
      }

    }

    # w[is.nan(w)] <- 0 # to avoid 0/0 in v

    # ++++++ #
    # M-Step #
    # ++++++ #

    # mu

    mu.old <- mu

    if(is.null(mu.TRUE)){
      den    <- matrix(rep(1/colSums(w),n),n,d,byrow=TRUE)
      mu     <- colSums(w*den*X)
      mu.rep <- matrix(rep(mu,n),n,d,byrow=TRUE)
    }

    # Sigma

    Sigma.old <- Sigma

    if(is.null(Sigma.TRUE)){
      Sigma     <- crossprod(sqrt(w)*(X-mu.rep))/n
      Sigma.rep <- matrix(rep(diag(Sigma),n),n,d,byrow=TRUE)
    }

    # theta

    theta.old <- theta

    if(algorithm == "ECME"){

      theta <- dsdist.ML.theta(X = X,
                               weights = NULL,
                               model = model,
                               common.theta = common.theta,
                               theta0 = theta,
                               mu = mu,
                               Sigma = Sigma,
                               method = method.theta,
                               iter.max = maxit,
                               reltol = tol,
                               trace = trace.theta)

    }

    if(algorithm == "EM"){

      if(!common.theta){

        for(h in 1:d)
          theta[h] <- n/(sum(w[,h])-n)

      }

      if(common.theta){

        theta.new <- n*d/(sum(w)-n*d)
        theta     <- rep(theta.new,d)

      }

    }

    # +++++++++++++++++++ #
    # Density Computation #
    # +++++++++++++++++++ #

    if(model == "sen")
      dens <- dDSSEN(X, mu = mu, Sigma = Sigma, theta = theta)
    if(model == "tin")
      dens <- dDSTIN(X, mu = mu, Sigma = Sigma, theta = theta)

    # ------------------------------------- #
    # Global - Observed-data log-likelihood #
    # ------------------------------------- #

    llvalues <- sum(log(dens))
    loglik   <- c(loglik,llvalues)

    iteration <- iteration + 1
    cat("*")

    if((loglik[iteration]-loglik[iteration-1]) < 0){

      warning("One of the units may be close to the estimated mean")

      iteration <- iteration - 1
      w     <- w.old
      mu    <- mu.old
      Sigma <- Sigma.old
      theta <- theta.old

      check <- 1

    }

    if(iteration == maxit | (loglik[iteration]-loglik[iteration-1]) < tol)
      check <- 1

  }

  # beepr::beep(4)

  # The EM-algorithm is finished

  # -------------------- #
  # Final Log-likelihood #
  # -------------------- #

  final.loglik <- loglik[iteration]

  # ------------------------------- #
  # Marginal Variances and kurtosis #
  # ------------------------------- #

  Dim.var <- Dim.kurt <- v.theta <- numeric(d)
  if(model=="sen"){

    v.theta  <- theta*exp(theta)*zipfR::Igamma(a = 0, x = theta, lower = FALSE)
    Dim.var  <- v.theta*diag(Sigma)
    Dim.kurt <- (1 - theta*exp(theta)*zipfR::Igamma(a = 0, x = theta, lower = FALSE))/(theta*exp(2*theta)*zipfR::Igamma(a = 0, x = theta, lower = FALSE)^2)*3

    num    <- expint::gammainc(a = (1/2 + 2), x = (delta[i,h]/2 + theta[h]))
    den    <- (delta[i,h]/2 + theta[h])*expint::gammainc(a = (1/2 + 1), x = (delta[i,h]/2 + theta[h]))
    w[i,h] <- num/den

    v.theta  <- theta*exp(theta)*zipfR::Igamma(a = 0, x = theta, lower = FALSE)
    Dim.var  <- v.theta*diag(Sigma)
    Dim.kurt <- (1 - theta*exp(theta)*zipfR::Igamma(a = 0, x = theta, lower = FALSE))/(theta*exp(2*theta)*zipfR::Igamma(a = 0, x = theta, lower = FALSE)^2)*3

  }
  if(model=="tin"){

    v.theta  <- -log(1-theta)/theta
    Dim.var  <- v.theta*diag(Sigma)
    k.theta  <- theta^2/((1-theta)*log(1-theta)^2)
    Dim.kurt <- k.theta*3

  }

  # ------------------------------------------------ #
  # Plot to check monotonicity of the log-likelihood #
  # ------------------------------------------------ #

  # if(plotll){
  #
  par(mai=c(0.84,0.8,0.012,0.004))
  par(las = 3)
  par(cex.axis=0.7)
  par(cex.lab=1.2)
  plot(1:(iteration-1), loglik[2:iteration], type="l",axes = FALSE, xlab="iterations", ylab="Objective function", lwd=2)
  axis(1, at = 1:(iteration-1), label = 1:(iteration-1))
  axis(2)
  box(col = "black")
  #
  # }

  # -------------------- #
  # Number of parameters #
  # -------------------- #

  if(!common.theta)
    npar <- d + d + d*(d-1)/2 + d
  if(common.theta)
    npar <- d + d + d*(d-1)/2 + 1

  # -------------------- #
  # Information criteria #
  # -------------------- #

  AIC <- 2*final.loglik - 2*npar
  BIC <- 2*final.loglik - npar*log(n)

  result <- list(
    X         = X,
    n         = n,
    d         = d,
    npar      = npar,
    mu.init     = mu.init,
    Sigma.init  = Sigma.init,
    theta.init  = theta.init,
    mu        = mu,
    Sigma     = Sigma,
    theta     = theta,
    w         = w,
    Dim.var   = Dim.var,
    Dim.kurt  = Dim.kurt,
    density   = dens,
    iter.stop = iteration,
    loglik    = final.loglik,
    AIC       = AIC,
    BIC       = BIC
    #call      = match.call()
  )
  #class(result) <- "MSEN"

  return(result)

}
