#' Weighted ML for the MSEN distribution: Direct approach.
#'
#' @param X Data matrix.
#' @param weights Vector of weights.
#' @param method The optimization method used by \code{\link[stats]{optim}}.
#' @param start.mu Initial mean vector.
#' @param start.Sigma Initial scalar matrix.
#' @param start.theta Initial inflation parameter.
#' @param maxit The maximum number of iterations.
#' @param reltol Relative convergence tolerance.
#' @param trace Non-negative integer. If positive, tracing information on the progress of the optimization is produced. Higher values may produce more tracing information (see \code{\link[stats]{optim}} for further details).
#' @param formula Method used to calculate the density: \code{"direct"}, \code{"indirect"}, \code{"series"}.
#'
#' @return A list with the following elements:
#' \item{loglik}{Log-likelihood at convergence.}
#' \item{AIC}{AIC.}
#' \item{BIC}{BIC.}
#' \item{npar}{Number of (free) parameters.}
#' \item{mu}{Mean vector.}
#' \item{U}{Upper triangular matrix from the Cholesky decomposition.}
#' \item{Sigma}{Scalar matrix.}
#' \item{theta}{Inflation parameter.}
#' @examples
#' d <- 2
#' set.seed(1234)
#' X <- rmsen(n = 100, mu = rep(0,d), Sigma = diag(d), theta = 1)$X
#' msen.WML(X = X)
#'
#' @export
dsdist.WML <- function(X,
                       weights = NULL,
                       model = "sen",
                       common.theta = FALSE,
                       method = "Nelder-Mead",
                       start.mu = NULL,
                       start.Sigma = NULL,
                       start.theta = NULL,
                       maxit = 5000,
                       reltol = 1e-15,
                       trace = 0){

  if(is.vector(X))
    X <- matrix(X, ncol = 1)
  if(is.data.frame(X))
    X <- as.matrix(X)
  if(any(is.na(X)))
    stop('No NAs allowed.')
  if(common.theta & length(start.theta)>1){
    start.theta <- start.theta[1]
  }

  n <- nrow(X)
  d <- ncol(X)

  # -------------------- #
  # Number of parameters #
  # -------------------- #

  if(!common.theta)
    npar <- d+d*(d+1)/2+d
  if(common.theta)
    npar <- d+d*(d+1)/2+1

  if(is.null(weights))
    weights <- rep(1,n)

  if(is.null(start.mu) | is.null(start.Sigma) | is.null(start.theta)){

     temp <- dsdist_WMM(X = X, model = model, weights = weights)

    #temp <- Rfast::mvnorm.mle(X)

  }
  if(is.null(start.mu)){
    start.mu <- temp$mu
  }
  if(is.null(start.Sigma)){
    #start.Sigma <- temp$sigma
     start.Sigma <- temp$Sigma
  }
  if(is.null(start.theta)){
    if(!common.theta)
      start.theta <- temp$theta
    if(common.theta)
      start.theta <- mean(temp$theta)
  }

  # initialization

  initial.values <- numeric(npar)

  initial.values[1:d] <- start.mu
  start.U <- base::chol(start.Sigma, pivot=FALSE)
  cont  <- 0
  for(i in 1:d)
    for(j in i:d){
      cont <- cont + 1
      initial.values[d+cont] <- start.U[i,j]
    }
  if(!common.theta){
    if(model == "sen")
      initial.values[(d+d*(d+1)/2+1):(d+d*(d+1)/2+d)] <- log(start.theta)
    if(model == "tin")
      initial.values[(d+d*(d+1)/2+1):(d+d*(d+1)/2+d)] <- log(start.theta/(1-start.theta))
  }
  if(common.theta){
    if(model == "sen")
      initial.values[d+d*(d+1)/2+1] <- log(start.theta)
    if(model == "tin")
      initial.values[d+d*(d+1)/2+1] <- log(start.theta/(1-start.theta))
  }

  # Objective function (pseudo log-likelihood)

  f <- function(par,X,weights,d,model,common.theta){

    # -- #
    # mu #
    # -- #

    mu <- array(0,c(d),dimnames=list(paste("X.",1:d,sep="")))
    for(i in 1:d)
      mu[i] <- par[i]

    # ----- #
    # Sigma #
    # ----- #

    U     <- array(0,c(d,d),dimnames=list(paste("X.",1:d,sep=""),paste("X.",1:d,sep="")))
    cont  <- 0
    for(i in 1:d)
      for(j in i:d){
        cont   <- cont + 1
        U[i,j] <- par[d+cont]
      }
    Sigma <- crossprod(x=U,y=U)

    # ----- #
    # theta #
    # ----- #

    if(!common.theta){
      if(model == "sen")
        theta <- exp(par[(d+d*(d+1)/2+1):(d+d*(d+1)/2+d)])
      if(model == "tin")
        theta <- exp(par[(d+d*(d+1)/2+1):(d+d*(d+1)/2+d)])/(1+exp(par[(d+d*(d+1)/2+1):(d+d*(d+1)/2+d)]))
    }
    if(common.theta){
      if(model == "sen")
        theta <- rep(exp(par[d+d*(d+1)/2+1]),d)
      if(model == "tin")
        theta <- rep(exp(par[d+d*(d+1)/2+1])/(1+exp(par[d+d*(d+1)/2+1])),d)
    }

    # -------------- #
    # -loglikelihood #
    # -------------- #

    if(model == "sen"){
      pll <- sum(weights*log(dDSSEN(X, mu = as.vector(mu), Sigma = Sigma, theta = theta)))
    }
    if(model == "tin"){
      pll <- sum(weights*log(dDSTIN(X, mu = as.vector(mu), Sigma = Sigma, theta = theta)))
    }

    return(pll)

  }

  res <- stats::optim(par = initial.values,
                      fn = f,
                      X = X,
                      weights = weights,
                      d = d,
                      model = model,
                      common.theta = common.theta,
                      method = method,
                      control=list(maxit = maxit, reltol = reltol, trace = trace, fnscale = -1))

  loglik  <- res$value
  est     <- res$par

  # -- #
  # mu #
  # -- #

  estmu <- array(0,c(d),dimnames=list(paste("X.",1:d,sep="")))
  for(i in 1:d)
    estmu[i] <- est[i]

  # ----- #
  # Sigma #
  # ----- #

  estU  <- array(0,c(d,d),dimnames=list(paste("X.",1:d,sep=""),paste("X.",1:d,sep="")))
  cont  <- 0
  for(i in 1:d)
    for(j in i:d){
      cont      <- cont + 1
      estU[i,j] <- est[d+cont]
    }
  estSigma <- crossprod(x=estU,y=estU)

  # ----- #
  # theta #
  # ----- #

  if(!common.theta){
    if(model == "sen")
      esttheta <- exp(est[(d+d*(d+1)/2+1):(d+d*(d+1)/2+d)])
    if(model == "tin")
      esttheta <- exp(est[(d+d*(d+1)/2+1):(d+d*(d+1)/2+d)])/(1+exp(est[(d+d*(d+1)/2+1):(d+d*(d+1)/2+d)]))
  }
  if(common.theta){
    if(model == "sen")
      esttheta <- rep(exp(est[d+d*(d+1)/2+1]),d)
    if(model == "tin")
      esttheta <- rep(exp(est[d+d*(d+1)/2+1])/(1+exp(est[d+d*(d+1)/2+1])),d)
  }

  # -------------------- #
  # Information criteria #
  # -------------------- #

  AIC <- 2*loglik - 2*npar
  BIC <- 2*loglik - npar*log(n)

  return(
    list(
      loglik = loglik,
      AIC    = AIC,
      BIC    = BIC,
      npar   = npar,
      mu     = estmu,
      U      = estU,
      Sigma  = estSigma,
      theta  = esttheta
    )
  )

}
