#' CM step for the estimation of theta in the ECME algorithm for DS distributions
#'
#' @param X Data matrix.
#' @param weights Vector of weights.
#' @param method The optimization method used by \code{\link[stats]{optim}}.
#' @param iter.max The maximum number of iterations.
#' @param reltol Relative convergence tolerance.
#' @param trace Non-negative integer. If positive, tracing information on the progress of the optimization is produced. Higher values may produce more tracing information (see \code{\link[stats]{optim}} for further details).
#'
#' @return Estimate for theta.
#'
#'
#'
#' @export
dsdist.ML.theta <- function(X,
                            weights = NULL,
                            model = "tin",
                            common.theta = FALSE,
                            theta0 = NULL,
                            mu,
                            Sigma = NULL,
                            method = "Nelder-Mead",
                            iter.max = 5000,
                            reltol = 1e-15,
                            trace = 0){

  if(is.vector(X))
    X <- matrix(X, ncol = 1)
  if(is.data.frame(X))
    X <- as.matrix(X)
  if(!is.matrix(X))
    X <- matrix(X, nrow = length(X), ncol = 1, dimnames = list(names(X), deparse(substitute(X))))
  if(!is.numeric(X))
    stop("numeric matrix/vector expected for X")
  if(any(is.na(X)))
    stop("No NAs allowed")
  if(common.theta & length(theta0)>1){
    theta0 <- theta0[1]
  }

  n <- nrow(X) # sample size
  d <- ncol(X) # number of variables

  if(is.null(weights))
    weights <- rep(1,n)

  if(!common.theta){

    if(is.null(theta0)){

      if(model=="tin")
        theta0 <- rep(0.01,d)
      if(model=="sen")
        theta0 <- rep(0.1,d)

    }

    if(model=="tin")
      initial.values <- log(theta0/(1-theta0))

    if(model=="sen")
      initial.values <- log(theta0)

    # objective function

    f <- function(par,weights,X,mu,Sigma,model){

      # ----- #
      # theta #
      # ----- #

      if(model=="tin")
        theta <- exp(par)/(1 + exp(par))

      if(model=="sen")
        theta <- exp(par)

      # -------------------- #
      # pseudo loglikelihood #
      # -------------------- #

      if(model=="tin")
        pll <- sum(weights*log(dDSTIN(x=X, mu=mu, Sigma=Sigma, theta=theta)))
      if(model=="sen")
        pll <- sum(weights*log(dDSSEN(x=X, mu=mu, Sigma=Sigma, theta=theta)))

      return(pll)

    }

    res <- optim(par = initial.values,
                 fn = f,
                 weights = weights,
                 X = X,
                 mu = mu,
                 Sigma = Sigma,
                 model = model,
                 method = method,
                 control = list(maxit = iter.max, reltol = reltol, trace = trace, fnscale = -1))

    f.obj <- res$value
    est   <- res$par

    if(model=="tin")
      theta <- exp(est)/(1 + exp(est))
    if(model=="sen")
      theta <- exp(est)

  }

  if(common.theta){

    if(model == "tin"){

      # objective function

      f <- function(par,d,weights,X,mu,Sigma){

        # ----- #
        # theta #
        # ----- #

        theta <- rep(par,d)

        # -------------------- #
        # pseudo loglikelihood #
        # -------------------- #

        if(model=="tin")
          pll <- sum(weights*log(dDSTIN(x=X,mu=mu,Sigma=Sigma,theta=theta)))
        if(model=="sen")
          pll <- sum(weights*log(dDSSEN(x=X,mu=mu,Sigma=Sigma,theta=theta)))

        return(pll)

      }

      res <- optimize(f = f,
                      d = d,
                      weights = weights,
                      X = X,
                      mu = mu,
                      Sigma = Sigma,
                      lower = 0,
                      upper = 1,
                      maximum = TRUE)

      f.obj <- res$objective
      theta <- rep(res$maximum,d)

    }

    if(model == "sen"){

      if(is.null(theta0))
        theta0 <- 0.1

      # objective function

      f <- function(par,d,weights,X,mu,Sigma){

        # ----- #
        # theta #
        # ----- #

        theta <- rep(par,d)

        # -------------------- #
        # pseudo loglikelihood #
        # -------------------- #

        if(model=="tin")
          pll <- -sum(weights*log(dDSTIN(x=X,mu=mu,Sigma=Sigma,theta=theta)))
        if(model=="sen")
          pll <- -sum(weights*log(dDSSEN(x=X,mu=mu,Sigma=Sigma,theta=theta)))

        return(pll)

      }

      res <- nlminb(start = theta0,
                    objective = f,
                    # control = list(),
                    d = d,
                    weights = weights,
                    X = X,
                    mu = mu,
                    Sigma = Sigma,
                    lower = 0,
                    upper = Inf)

      f.obj <- res$objective
      theta <- rep(res$par,d)

    }

  }

  return(theta)

}
