#' Method of weighted moments.
#'
#' @param X Data matrix.
#' @param model marginal distribution considered. Possible values are: \code{"sen"} or \code{"tin"}
#' @param weights Vector of mixing weights.
#'
#' @return A list with the following elements:
#' \item{d}{Dimension.}
#' \item{mu}{Mean vector.}
#' \item{Var}{Covariance matrix.}
#' \item{Kurt}{Kurtosis.}
#' \item{Sigma}{Scalar matrix.}
#' \item{theta}{Tailedness parameter.}
#' @examples
#' d <- 2
#' set.seed(1234)
#' X <- rDSSEN(100,mu=rep(0,d),Sigma=diag(d),theta=rep(0.95,d))$X
#' dsdist_WMM(X=X)
#'
#' @export
dsdist_WMM <- function(X, model = "sen", weights = NULL){

  if(is.vector(X))
    X <- matrix(X,ncol=1)
  if(is.data.frame(X))
    X <- as.matrix(X)
  if(any(is.na(X)))
    stop('No NAs allowed.')

  d <- ncol(X)
  n <- nrow(X)

  if(is.null(weights))
    weights <- rep(1,n)

  mu   <- apply(X,2,stats::weighted.mean,w=weights)
  Var  <- stats::cov.wt(x=X, wt = weights, cor = FALSE, center = TRUE, method = "ML")$cov

  Kurt <- NULL

  for(i in 1:d){
    Kurt[i] <- max(mean((X[,i]-mu[i])^4/Var[i,i]^2),3)
  }

  if(model == "tin"){

    # co-variance factor

    cv <- function(theta1, theta2){
      4*1/(theta1*theta2)*(1-sqrt(1-theta1))*(1-sqrt(1-theta2))
    }

    # variance factor

    v <- function(theta){
      -log(1-theta)/theta
    }

    # kurtosis factor

    k <- function(theta){
      theta^2/((1-theta)*log(1-theta)^2)
    }

  }

  if(model == "sen"){

    # co-variance factor

    cv <- function(theta1, theta2){
      pi*exp(theta1)*sqrt(theta1)*pracma::erfc(sqrt(theta1))*exp(theta2)*sqrt(theta2)*pracma::erfc(sqrt(theta2))
    }

    # variance factor

    v <- function(theta){
      theta*exp(theta)*expint::gammainc(a = 0, x = theta)
    }

    # kurtosis factor

    k <- function(theta){
      (1 - theta*exp(theta)*expint::gammainc(a = 0, x = theta))/(theta*exp(2*theta)*expint::gammainc(a = 0, x = theta)^2)
    }

  }

  # objective function

  fobj <- function(theta,ktheta){
    abs(k(theta=theta) - ktheta)^2
  }

  ktheta <- theta <- NULL

  # Search for theta

  if(model == "tin"){

    for(i in 1:d){

      ktheta[i] <- Kurt[i]/3

      if(ktheta[i]==1)
        theta[i] <- 0.5

      if(ktheta[i]>1)
        theta[i] <- stats::optimize(f=fobj, interval=c(0, 1), ktheta[i])$minimum
    }

  }

  if(model == "sen"){

    for(i in 1:d){

      ktheta[i] <- Kurt[i]/3

      if(ktheta[i]==1)
        theta[i] <- 100

      if(ktheta[i]>1)
        theta[i] <- stats::optimize(f=fobj, interval=c(0, 100), ktheta[i])$minimum
    }

  }

  # Scale Matrix

  if(d > 1)
    Sigma0 <- outer(theta,theta,cv)*(1-diag(d))
  if(d == 1)
    Sigma0 <- 0

  #Sigma0 <- outer(theta,theta,cv)*(1-diag(d))
  #print(Sigma0)
  Sigma1 <- Sigma0 + diag(v(theta))

  #Sigma  <- ifelse(d>1, Var/Sigma1 , as.numeric(Var)/Sigma1)
  #print(str(Var))

  if(d > 1)
    Sigma <- Var/Sigma1
  if(d == 1)
    Sigma <- as.numeric(Var)/Sigma1

  return(
    list(
      d     = d,
      mu    = mu,
      Var   = Var,
      Kurt  = Kurt,
      Sigma = Sigma,
      theta = theta
    )
  )

}
