#' Generate random numbers from the DSSEN distribution.
#'
#' @param n Number of points to be generated.
#' @param mu Mean vector.
#' @param Sigma Scale matrix.
#' @param theta Vector of tailedness parameters for each dimension.
#' @param norm.dens Method used to calculate the density of the multivariate normal distribution: \code{"dmnorm"} (see \code{\link[mnormt]{dmnorm}}) or \code{"dmvnorm"}  (see \code{\link[mvtnorm]{Mvnorm}}).
#'
#' @return A list with the following elements:
#' \item{X}{Data matrix.}
#' \item{w}{Vector of mixing weights.}
#' @examples
#' d <- 3
#' rDSSEN(10,mu=rep(0,d),Sigma=diag(d),theta=rep(1,d))
#'
#' @export
rDSSEN <- function(n, mu = rep(0,d), Sigma = NULL, theta = rep(500,d), norm.dens = "dmnorm"){

  # Lambda is a vector with diagonal elements

  if(min(theta) < 0)
    stop("each element in theta must be greater than, or equal to, 0")

  if(is.vector(mu))
    d <- length(mu)
  if(!is.vector(mu))
    d <- 1

  X <- w <- array(0,c(n,d),dimnames=list(1:n,paste("X.",1:d,sep="")))
  for(h in 1:d)
    w[,h] <- (1 + stats::rexp(n = n, theta[h]))

  DeltaRS <- SigmaMod <- array(0,c(n,d,d),dimnames=list(1:n,paste("X.",1:d,sep=""),paste("X.",1:d,sep="")))
  for(i in 1:n){
    if(d > 1)
      DeltaRS[i,,] <- diag(1/sqrt(w[i,]))
    if(d == 1)
      DeltaRS[i,,] <- 1/sqrt(w[i,])
    SigmaMod[i,,] <- DeltaRS[i,,] %*% Sigma %*% DeltaRS[i,,]
    if(norm.dens == "dmnorm")
      X[i,] <- mnormt::rmnorm(n = 1, mean = mu, varcov = SigmaMod[i,,])
    if(norm.dens == "dmvnorm")
      X[i,] <- mvtnorm::rmvnorm(n = 1, mean = mu, sigma = SigmaMod[i,,])
  }

  return(list(X = X, w = w))

}
